//
//     WARNING
//       This code is completely open and available to be modified and changed
//       in any way and by anybody. It is yours once you start to change it.
//       Don't try to patent your derivative code and sue us for infringement.
//       We were here first!
//
//       The reason it is provided this way is for your education and to illustrate
//       some ideas. It can damage your Arduino if you are not extremely careful.
//       No guarantee is offered and no risk is assumed by the code author.
//       All the risk is on you even if you follow the commments provided in the
//       source code or elsewhere. This is powerful code but you must know it well.
//
//       This code implements the Servo Control Module as imagined by Robot-DHQ.
//       It has many undocumented features and has not been optimized. You can learn
//       a lot by studying it or you can use it according to the documentation and
//       not read any further. Notes within the source code will give you some ideas
//       for using some of the undocumented features.
//
//       Robot-DHQ is the Robot Design Headquarters
//
//       Consider it a "work in progress". The Servo Control Module is described in
//       this document: D:\Projects\Arduino\Smart Modules\Servo Control Module.odt
//
//
// how to avoid destroying the Arduino Nano
// 1   the Arduino will not be detected over the USB interface if the Schottky diode
//     has been destroyed. This can happen if the +5 volt supply is accidentally
//     shorted to ground. The diode can only handle about 0.5 amps before it blows out.
//     This is therefore the first thing to suspect if the serial connection drops.
//     Therefore, always check for shorts before applying power
//
// 2   input pins are relatively safe to implement as long as they are not
//     connected to a power source with a voltage greater or less than the
//     hardware limitations, usually VCC and GROUND. The S1025 command is
//     provided to produce an input only state.
//

#include <SoftwareSerial.h>

// it seems that pin 1 (RX) on the Arduino Nano is not easily driven high and low by
// another Arduino Nano and I did not find an easy solution online so I have added
// the SoftwareSerial library to receive serial commands
// commands from the two serial ports are processed the same ... see the code!
// Arduino pin D12 is LED 6 on the LED board when the board is attached at position D
// Arduino pin D10 is LED 4 on the LED board when the board is attached at position D
// Arduino pin D6 is LED 8 on the LED board when the board is attached at position D
// Arduino pin D6 is also position 2 on the servo board when the board is attached at position C
SoftwareSerial gtSerial(6, 12); // RX is Arduino pin D6, TX is Arduino pin D12

// port allocations, etc.
const int NANO_05_D2 = 2;   // Socket C Pin 8, Socket E Pin 4
const int NANO_06_D3 = 3;   // Socket C Pin 5, Socket E Pin 5
const int NANO_07_D4 = 4;   // Socket C Pin 4, Socket E Pin 8
const int NANO_08_D5 = 5;   // Socket C Pin 1, Socket E Pin 7
const int NANO_09_D6 = 6;   // Socket C Pin 2, Socket B Pin 4, Socket D Pin 10
const int NANO_10_D7 = 7;   // Socket B Pin 5, Socket D Pin 9
const int NANO_11_D8 = 8;   // Socket B Pin 8, Socket D Pin 7
const int NANO_12_D9 = 9;   // Socket B Pin 7, Socket D Pin 1
const int NANO_13_D10 = 10; // Socket D Pin 5
const int NANO_14_D11 = 11; // Socket D Pin 2
const int NANO_15_D12 = 12; // Socket D Pin 8
const int NANO_16_D13 = 13; // Socket D Pin 4
const int NANO_19_A0 = A0;  // Socket A Pin 1, Socket E Pin 2
const int NANO_20_A1 = A1;  // Socket A Pin 4, Socket E Pin 1
const int NANO_21_A2 = A2;  // Socket A Pin 2, Socket E Pin 10
const int NANO_22_A3 = A3;  // Socket A Pin 10, Socket E Pin 9
const int NANO_23_A4 = A4;  // Socket A Pin 8, Socket B Pin 9
const int NANO_24_A5 = A5;  // Socket A Pin 9, Socket B Pin 1
const int NANO_25_A6 = A6;  // Socket A Pin 7, Socket B Pin 10
const int NANO_26_A7 = A7;  // Socket A Pin 5, Socket B Pin 2

// these variables affect the configuration settings at startup
//   timing parameters
//     the timing defaults can be adjusted by changing these parameters here
const int slow_loop_size = 50;
// fast_loop_size can be adjusted to implement a pseudo-millisecond timer
const unsigned long fast_loop_size = 250;
// default configuration of LED and PWM variables
const int default_mid_position = 1500; // default servo pulse width in microseconds
const int default_max_position = 2012; // default max servo pulse width in microseconds
const int default_min_position = 988;  // default min servo pulse width in microseconds
const int default_slew_rate = 10;
const int pwm_idx_cv = 0xB;
// end of configuration constants

// user interface variables
//     the user documentation may be consulted for the use of these variables
//     these variables control how the various commands are used
int remote_mode = 0;         // operating mode
int analog_pin = 0;  // the current analog pin, range 0-7
int digital_pin = 0; // the current digital pin, range 0-11, add 2 for display
int tracker_on = 0; // only track the current digital (PWM) pin
int tracker_tracking = 0;

// timing variables
unsigned long startMicros = 0;
unsigned long servoStart = 0;
unsigned long currentMicros = 0;
// pulse_loop_size controls how often a PWM pulse is sent, a.k.a. the PWM frequency
// it is the length in microseconds of the "window" in which the PWM pulses are high
// it is only tested after the longest PWM pulse has been sent therefore
// it is effectually meaningless if shorter than the longest PWM pulse
// it can be adjusted from 0 to 30,000 microseconds using the serial interface
// to achieve the "standard" rate of 50 times per second this should be 20,000 microseconds
// many servos are quite flexible and this rate can be adjusted to achieve various effects
// jitter and jerkiness can sometimes be reduced by increasing this value and pulsing less often
// a value of 10,000 is a period of approx 10,000 microseconds, one pulse per 10 millisecond or 100 pulses per second
// a value of 25,000 is a period of approx 25,000 microseconds, one pulse per 25 millisecond or 40 pulses per second
// a value of 2,500 is a period of approx 2,500 microseconds, one pulse per 2.5 millisecond or 400 pulses per second
// these values are approximate because CPU overhead for serial IO, etc. is not included
int pulse_loop_size = 4000;

// track_loop_size controls how often PWM position and target data is transmitted
// a value of 250 is approx one transmission per 250 millisecond / 4 times per second
// the tracker can be used to estimate the frequecy of the fast and slow software loops
// this can be one way to adjust fast_loop_size and slow_loop_size
int track_loop_size = 250;

// led_loop_size controls how often LEDs are handled when NOT simultaneously handling PWM
int led_loop_size = 25;

// the pin_ array variables configure and control the digital IO pins
//   pin_function indicates the current function:
//     0: raw two-state input
//     1: debounced input
//     2: LED output
//     3: PWM output
//     4: ARP Arduino PWM output (future)
//   pin_control_src indicates the type of source to use to control the digital output
//     0: target or brightness is set using remote commands
//     1: target or brightness is set based on two-state signal or input
//     2: target or brightness is set based on push button or trigger
//     3: target or brightness is calculated from an analog input
//     4: target or brightness is set based on analog high/low state
//     5: target or brightness is set based on analog high/low transition or trigger
//   pin_parm_src indicates where to find the parameters to use to control the digital output
//     0-11: a digital pin if pin_control_src is 1 or 2
//     0-7: an analog input if pin_control_src is 3, 4 or 5
//   pin_locked indicates whether the function can be changed on command while running
//     a value of 1 indicates that function cannot be changed by a remote command
//   pin_io is 0 if configured for input, 1 if configured for output
//   pin_state is the raw digital input
//   pin_last is used to debounce a pushbutton switch
//   pin_now is the digital input raw or debounced
//   pin_d_prior_state is used to identify the transition
//   pin_led_off_time is the off duration (if LED)
//   pin_led_on_time is the on duration (if LED)
//   pin_servo_pos is the width of the pulse in microseconds (if PWM)
//   pin_servo_target is the target pulse width (if PWM)
//   pin_servo_max is the maximum pulse width (LED or PWM)
//   pin_servo_min is the minimum pulse width (LED or PWM)
//   pin_servo_range is the pulse width range (LED or PWM)
//   pin_slew_rate controls the servo slew rate (if PWM)
//   pin_counter measures pulse width (LED or PWM)
//   pin_servo_done marks the end of the pulse (if PWM)
//   pin_reverse_out reverses the sense of the output (LED or PWM)
//   pin_d_tran_event flags a transition event of a digital input
//   pin_d_toggle_state is used by the automation cross bar logic
//   pin_parm_1 holds a value that can dim an LED or set a servo
//   pin_parm_2 holds a value that can dim an LED or set a servo

int pin_function[12];
int pin_io[12];
int pin_control_src[12];
int pin_parm_src[12];
int pin_locked[12];
int pin_state[12];
int pin_last[12];
int pin_now[12];
int pin_led_off_time[12];
int pin_led_on_time[12];
int pin_servo_pos[12];
int pin_servo_target[12];
int pin_servo_max[12];
int pin_servo_min[12];
int pin_servo_range[12];
int pin_slew_rate[12];
int pin_counter[12];
int pin_servo_done[12];
int pin_reverse_out[12];
int pin_d_prior_state[12];
int pin_d_tran_event[12];
int pin_d_toggle_state[12];
int pin_parm_1[12];
int pin_parm_2[12];

// analog input changes are ignored if the change is small
// pin implementation variables
//   pin_A*_analog_in is the raw analog input
//   pin_A*_min_change is the change needed
//   pin_A*_last_value is the reference value
//   pin_A*_changed indicates value has changed
//   pin_a_parm_1 holds a value that can dim an LED or set a servo
//   pin_a_parm_2 holds a value that can dim an LED or set a servo

int analog_in_pin[8];
int analog_in_value[8];
int analog_in_min[8];
int analog_in_max[8];
int analog_in_range[8];
int pin_a_now[8];
int pin_a_prior_state[8];
int pin_a_tran_event[8];
int pin_a_toggle_state[8];
int pin_a_parm_1[8];
int pin_a_parm_2[8];
unsigned long analog_in_scaled[8];
//int pin_A0_min_change = 5;
//int pin_A0_last_value = 0;
//int pin_A0_changed = 0;

// temporary variables
unsigned long tmp_val_long = 0;
int tmp_val_int = 0;

// indexes and counters
int analog_read_idx = 0;
int pin_out = 0;
int pin_in = 0;
int pin_idx;
int pwm_idx = 0;
int led_idx = 0;
int ain_idx;
int slow_counter = 0;
int track_counter = 0;
int led_loop_counter = 0;

// servo PWM controls
int sending_pwm_pulses = 0;
int pulses_are_done_flag;

// serial input eval
int i;
int num_chars = 0;
const byte numChars = 32;
char receivedChars[numChars];
boolean serial_available = false;
boolean newData = false;

void setup() {
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    setPinIO(pin_idx, 1); // set digital pins to input
  }

  gtSerial.begin(57600);  // software serial port
  Serial.begin(57600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  Serial.println("Servo Module");
  blinkLED6();
  initialize_analog_arrays();
  initialize_digital_arrays();
  // change startup mode here!
  initialize_demo_mode();
  //initialize_input_only_mode();
}

void initialize_analog_arrays() {
  analog_in_pin[0] = A0;
  analog_in_pin[1] = A1;
  analog_in_pin[2] = A2;
  analog_in_pin[3] = A3;
  analog_in_pin[4] = A4;
  analog_in_pin[5] = A5;
  analog_in_pin[6] = A6;
  analog_in_pin[7] = A7;
  for (ain_idx = 0; ain_idx < 8; ain_idx++) {
    analog_in_value   [ain_idx] = 0;
    analog_in_min     [ain_idx] = 0;
    analog_in_max     [ain_idx] = 1023;
    pin_a_now         [ain_idx] = 0;
    pin_a_prior_state [ain_idx] = 0;
    pin_a_tran_event  [ain_idx] = 0;
    pin_a_toggle_state[ain_idx] = 0;
    pin_a_parm_1      [ain_idx] = 0;
    pin_a_parm_2      [ain_idx] = 0;
    set_analog_range(ain_idx);
  }
}

void initialize_digital_arrays() {
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    pin_state          [pin_idx] = 1;
    pin_last           [pin_idx] = 0;
    pin_now            [pin_idx] = 0;
    pin_servo_pos      [pin_idx] = default_mid_position;
    set_servo_max(pin_idx, default_max_position);
    set_servo_min(pin_idx, default_min_position);
    pin_slew_rate      [pin_idx] = default_slew_rate;
    pin_counter        [pin_idx] = 0;
    pin_servo_done     [pin_idx] = 1;
    pin_reverse_out    [pin_idx] = 0;
    pin_d_tran_event   [pin_idx] = 0;
    pin_d_prior_state  [pin_idx] = 0;
    pin_d_toggle_state [pin_idx] = 0;
    pin_locked         [pin_idx] = 1;
    set_parm_1(pin_idx, pin_servo_range [pin_idx]);
    set_parm_2(pin_idx, 0);
    set_led_times(pin_idx, pin_idx, 0);
    set_target(pin_idx, default_mid_position);
  }
}

void loop() {
  currentMicros = micros();
  if (sending_pwm_pulses > 0) {
    // skip other processing such as serial input processing while pulses are being timed
    // this inner (fast) loop processes repeatedly until the PWM pulses are completely sent
    // for more refer to pulse_loop_size and servoStart
    if (3 == pin_function[pwm_idx]) { // PWM pins only
      if (pin_servo_done[pwm_idx] < 1) {
        if (currentMicros - servoStart >= pin_servo_pos[pwm_idx]) {
          digitalWrite(pwm_idx + 2, LOW);
          pin_servo_done[pwm_idx] = 1;
        } else {
          pulses_are_done_flag = 0; // strike the flag
        }
      }
    }
    pwm_idx++;
    if (pwm_idx > pwm_idx_cv) {
      pwm_idx = 0;
      if (pulses_are_done_flag > 0) { // flag still up?
        sending_pwm_pulses = 0; // all pulses are done!
      } else {
        pulses_are_done_flag = 1; // reset the flag
      }
      handle_led_on_off();
    }
  } else {
    recvWithStartEndMarkers();
    processCommand();
    led_loop_counter++;
    if (led_loop_counter > led_loop_size) {
      led_loop_counter = 0;
      handle_led_on_off();
    }
    if (currentMicros - startMicros > fast_loop_size) {
      slowLoop(); // about once per millisecond or 1000 times per second
      startMicros = currentMicros;
    }
    if (currentMicros - servoStart > pulse_loop_size) {
      sending_pwm_pulses = 1;
      servoStart = currentMicros;
      pulses_are_done_flag = 1; // raise the flag
      for (int i = 0; i < 12; i++) {
        if (3 == pin_function[i]) { // PWM pins only
          pin_servo_pos[i] = adjust_servo_position(pin_servo_pos[i], pin_servo_target[i], pin_slew_rate[i]);
          digitalWrite(i + 2, HIGH);
          pin_servo_done[i] = 0;
          pulses_are_done_flag = 0; // strike the flag
        }
      }
    }
  }
}

void slowLoop() { // about once per millisecond
  slow_counter++;
  if (slow_counter > slow_loop_size) {
    slow_counter = 0;
    // these actions are done about once per 50 millisecond or 20 times per second
    digital_reads();
    analog_read_and_scale();
    if (remote_mode < 1) {
      set_output_parms();
    }
    process_input_level_transitions();
  }

  if (tracker_on > 0) {
    track_counter++;
    if (track_counter > track_loop_size) {
      track_counter = 0;
      // when tracker is on this is done about every 250 millisecond or 4 times per second
      if (tracker_tracking > 0) {
        report_track();
        if (pin_servo_pos[digital_pin] == pin_servo_target[digital_pin]) {
          tracker_tracking = 0;
        }
      } else {
        if (pin_servo_pos[digital_pin] != pin_servo_target[digital_pin]) {
          tracker_tracking = 1;
        }
      }
    }
  }
}

void report_track() {
  print_user_pin();
  Serial.print(F("  "));
  print_num(pin_servo_pos[digital_pin]);
  Serial.print(F("   "));
  print_num(pin_servo_target[digital_pin]);
  Serial.println(F("   "));
}

// this function includes switch debounce logic
void digital_reads() {
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    pin_state[pin_idx] = digitalRead(pin_idx + 2);
    switch (pin_function[pin_idx]) {
      case 0: // don't debounce
        pin_now[pin_idx] = pin_state[pin_idx];
        break;
      case 1: // debounce
        if (pin_state[pin_idx] == pin_last[pin_idx]) {
          pin_now[pin_idx] = pin_state[pin_idx];
        }
        if (pin_d_prior_state[pin_idx] != pin_now[pin_idx]) {
          pin_d_prior_state[pin_idx] = pin_now[pin_idx];
          if (pin_d_prior_state[pin_idx] < 1) {
            toggle_d_toggle_state();
            pin_d_tran_event[pin_idx] = 12; // button press event
            Serial.println(F("d pin tran event   "));
          }
        }
        break;
    }
    pin_last[pin_idx] = pin_state[pin_idx];
  }
}

// this function shifts and scales the analog input readings
// the results are stored in the analog_in_scaled array
// the results may be used later to compute PWM pulse widths
//
// this was done initially for the following reason
// for an unknown reason one of my four pot inputs was strange
// the values from three of the analog reads ranged from 0 to 1023
// the fourth ranged from 0 to 529
// the min and max readings are set using the user interface
// the range is computed from these settings (range = max - min)
// the analog readings are multiplied by 8192 and divided by the range
// this is used later to compute the length of the PWM pulse
void analog_read_and_scale() {
  analog_in_value[analog_read_idx] = analogRead(analog_in_pin[analog_read_idx]);
  tmp_val_long = analog_in_value[analog_read_idx];
  if (tmp_val_long > 511) {
    pin_a_now[analog_read_idx] = 1;
  } else {
    pin_a_now[analog_read_idx] = 0;
  }
  if (pin_a_prior_state[analog_read_idx] != pin_a_now[analog_read_idx]) {
    pin_a_prior_state[analog_read_idx] = pin_a_now[analog_read_idx];
    if (pin_a_prior_state[analog_read_idx] < 1) {
      toggle_a_toggle_state(analog_read_idx);
      pin_a_tran_event[analog_read_idx] = 12; // button press event
    }
  }
  if (tmp_val_long > analog_in_max[analog_read_idx]) {
    tmp_val_long = analog_in_max[analog_read_idx];
  }
  if (tmp_val_long < analog_in_min[analog_read_idx]) {
    tmp_val_long = analog_in_min[analog_read_idx];
  }
  tmp_val_long -= analog_in_min[analog_read_idx];
  tmp_val_long *= 8192; // scale up to do integer math
  analog_in_scaled[analog_read_idx] = tmp_val_long / analog_in_range[analog_read_idx];
  // analog_in_scaled is the numerator of a fraction
  // the denominator of the fraction is 8192
  // the fraction is the portion of the analog range as determined by the user
  analog_read_idx++;
  if (analog_read_idx > 7) {
    analog_read_idx = 0;
  }
}

// this function sets various parameters that control the LED or PWM output
// the three major control setting are as follows
// 1. the output pin-function can be LED (2)or PWM (3) or ARO (4)
// 2. the output can be normal or reversed (pin_reverse_out 0/1)
// 3. the output is based on one of six control settings (pin_control_src valid values are 0 through 5)
//    (0) a value in memory set remotely
//        set the value in pin_target, etc.
//    (1) digital input
//        pin_parm_src indicates which digital input to be used
//        use the value in pin_parm_1 if pin_now is 0
//        use the value in pin_parm_2 if pin_now is 1
//    (2) momentary digital input toggles between two values
//        pin_parm_src indicates which digital input to be used
//        the alternate values are in pin_parm_1 and pin_parm_2
//    (3) pot input
//        pin_parm_src indicates which analog input to be used
//        analog min and max settings are used to preprocess the input
//        analog_in_scaled is used to control the output
//    (4) analog input value
//        pin_parm_src indicates which analog input to be used
//        use the value in pin_a_parm_1 if pin_a_now is 0
//        use the value in pin_a_parm_2 if pin_a_now is 1
//    (5) analog value transition input toggles between two values
//        pin_parm_src indicates which analog input to be used
//        the alternate values are in pin_a_parm_1 and pin_a_parm_2
void set_output_parms() {
  if (2 == pin_function[pin_out]) { // LED
    set_led_on_off_times();
  } else if (3 == pin_function[pin_out]) { // PWM
    set_servo_target();
  }
  pin_out++;
  if (pin_out > 11) {
    pin_out = 0;
  }
}

void set_led_on_off_times() {
  switch (pin_control_src[pin_out]) {
    case 1: // TWO-STATE:  act directly on pin_now state input
      if (pin_now[pin_parm_src[pin_out]] > 0) {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_parm_1[pin_parm_src[pin_out]]);
      } else {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_parm_2[pin_parm_src[pin_out]]);
      }
      break;
    case 2:  // TWO-STATE:  act directly on pin_now state input
      if (pin_d_toggle_state[pin_parm_src[pin_out]] > 0) {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_parm_1[pin_parm_src[pin_out]]);
      } else {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_parm_2[pin_parm_src[pin_out]]);
      }
      break;
    case 3: // POT
      tmp_val_long = analog_in_scaled[pin_parm_src[pin_out]] * pin_servo_range[pin_out];
      tmp_val_int = tmp_val_long / 8192;
      set_led_times(pin_out, pin_out, tmp_val_int);
      break;
    case 4: // TWO-STATE:  act directly on analog pin state input
      if (pin_a_now[pin_parm_src[pin_out]] > 0) {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_a_parm_1[pin_parm_src[pin_out]]);
      } else {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_a_parm_2[pin_parm_src[pin_out]]);
      }
      break;
    case 5: // TWO-STATE:  act directly on analog pin state toggle input
      if (pin_a_toggle_state[pin_parm_src[pin_out]] > 0) {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_a_parm_1[pin_parm_src[pin_out]]);
      } else {
        set_led_times(pin_out, pin_parm_src[pin_out], pin_a_parm_2[pin_parm_src[pin_out]]);
      }
      break;
  }
}

void set_servo_target() {
  switch (pin_control_src[pin_out]) {
    case 1: // TWO-STATE:  act directly on pin_now state input
      if (pin_now[pin_parm_src[pin_out]] > 0) {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_parm_1[pin_parm_src[pin_out]]));
      } else {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_parm_2[pin_parm_src[pin_out]]));
      }
      break;
    case 2: // TRIGGER SET:  act on push button press and release
      if (pin_d_toggle_state[pin_parm_src[pin_out]] > 0) {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_parm_1[pin_parm_src[pin_out]]));
      } else {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_parm_2[pin_parm_src[pin_out]]));
      }
      break;
    case 3: // POT
      // NORMAL target = servo_min + (servo_range * analog_in_scaled / divisor)
      // REVERSE target = servo_max - (servo_range * analog_in_scaled / divisor)
      // the goal is a target value that matches the pulse width characteristics of the attached servo
      // in the typical case, the result will range from 1000 to 2000 microseconds
      tmp_val_long = analog_in_scaled[pin_parm_src[pin_out]] * pin_servo_range[pin_out];
      tmp_val_int = tmp_val_long / 8192;
      set_target(pin_out, pin_servo_min[pin_out] + norm_or_reverse(pin_out, tmp_val_int));
      break;
    case 4: // TWO-STATE:  act directly on analog pin state input
      if (pin_a_now[pin_parm_src[pin_out]] > 0) {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_a_parm_1[pin_parm_src[pin_out]]));
      } else {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_a_parm_2[pin_parm_src[pin_out]]));
      }
      break;
    case 5: // TWO-STATE:  act directly on analog pin state toggle input
      if (pin_a_toggle_state[pin_parm_src[pin_out]] > 0) {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_a_parm_1[pin_parm_src[pin_out]]));
      } else {
        set_target(pin_out, pin_servo_min[pin_out] +
                   norm_or_reverse(pin_out, pin_a_parm_2[pin_parm_src[pin_out]]));
      }
      break;
  }
}

// digital input level transitions may have multiple effects
// all these effects must be processed before pin_tran_event is reduced to zero
// these effects may be processed here or elsewhere
// multiple servo positions could be affected for example
// servo positioning is done elsewhere
// after this is done the code here could send a serial command or take some other action
// this code is entered 12 times per cycle as each digital pin is processed
// that is why we count down until pin_tran_event is reduced to zero
void process_input_level_transitions() {
  for (pin_in = 0; pin_in < 8; pin_in++) {
    if (pin_a_tran_event[pin_in] > 0) {
      pin_a_tran_event[pin_in]--;
      if (pin_a_tran_event[pin_in] == 0) {
        switch (pin_in) {
          case 1: // button by S3 toggles LED6
            if (pin_a_toggle_state[1] < 1) {
              set_led_times(10, 10, 0);
            } else {
              set_led_times(10, 10, pin_servo_range[10]);
            }
            report_a_button_press();
            break;
          case 4: // button by S1 toggles LED2
            if (pin_a_toggle_state[4] < 1) {
              set_led_times(9, 9, 0);
            } else {
              set_led_times(9, 9, pin_servo_range[9]);
            }
            report_a_button_press();
            break;
          case 6: // button by S2 toggles LED4
            if (pin_a_toggle_state[6] < 1) {
              set_led_times(8, 8, 0);
            } else {
              set_led_times(8, 8, pin_servo_range[8]);
            }
            report_a_button_press();
            break;
          case 7: // button by S4 toggles LED8
            if (pin_a_toggle_state[7] < 1) {
              set_led_times(4, 4, 0);
            } else {
              set_led_times(4, 4, pin_servo_range[4]);
            }
            report_a_button_press();
            break;
        }
      }
    }
  }
  for (pin_in = 0; pin_in < 12; pin_in++) {
    if (pin_d_tran_event[pin_in] > 0) {
      pin_d_tran_event[pin_in]--;
      if (pin_d_tran_event[pin_in] == 0) {
        switch (pin_in) {
          case 90: // no such case for pin_in
            toggle_remote_mode();
            report_d_button_press();
            break;
        }
      }
    }
  }
}

// demo application function - change mode to local or remote
void report_a_button_press() { // application specific code
  Serial.print(F("analog pin_in "));
  Serial.print(pin_in);
  Serial.println();
}

// demo application function - change mode to local or remote
void report_d_button_press() { // application specific code
  Serial.print(F("D"));
  Serial.print(pin_in + 2);
  Serial.print(F(": Button ("));
  Serial.print(pin_in);
  Serial.println(") pressed.");
}

// demo application function - change mode to local or remote
void toggle_remote_mode() { // application specific code
  if (remote_mode > 0) {
    startLocalMode();
  } else {
    startRemoteMode();
  }
}

// the starter code for this section came from forum.arduino.cc
void recvWithStartEndMarkers() {
  static boolean recvInProgress = false;
  static byte ndx = 0;
  char startMarker = 'S';
  char endMarkerLF = 10;
  char endMarkerCR = 13;
  char rc;

  if (Serial.available() > 0 && newData == false) {
    rc = Serial.read();
    serial_available = true;
  }
  if (gtSerial.available() > 0 && newData == false) {
    rc = gtSerial.read();
    serial_available = true;
  }
  if (serial_available) {
    //Serial.print("rc=");
    //Serial.print(rc);
    //Serial.println("<<<");
    serial_available = false;
    if (recvInProgress == true) {
      if (rc == endMarkerLF) {
        receivedChars[ndx] = '\0'; // terminate the string
        recvInProgress = false;
        ndx = 0;
        newData = true;
      } else {
        if (rc == endMarkerCR) {
          receivedChars[ndx] = '\0'; // terminate the string
          recvInProgress = false;
          ndx = 0;
          newData = true;
        } else {
          receivedChars[ndx] = rc;
          ndx++;
          if (ndx >= numChars) {
            ndx = numChars - 1;
          }
        }
      }
    } else if (rc == startMarker) {
      recvInProgress = true;
    }
  }
}

void processCommand() {
  int command_number = 0;
  int num_chars = 0;
  if (newData == true) {
    // if you are reading this and are looking for a place to start, try this!
    // start the Arduino Serial Monitor in the Arduino IDE and
    // send S1999 to the Nano, be sure to use capital S
    // then try S1998, S1997, S1996, etc.
    // commands that return data start with S1999 and count down
    // any command not starting with S will be ignored
    // many other input character sequences are also ignored
    // this is intentional so take care what you do
    // silent commands start with S1001 and count up
    // the character sequences that return data are tightly restricted so
    // you may want to temporarily uncomment this Serial.print and Serial.println
    //    Serial.print("characters received = ");
    //    Serial.println(receivedChars);
    newData = false;
    for (int i = 0; i < 32; i++) {
      if (receivedChars[i] == '\0') {
        num_chars = i;
        i = 32;
      }
    }
    if (receivedChars[0] != '1') {
      invalidModuleCode(); // the first two characters of the command must be "S1"
    } else { //
      if (receivedChars[1] >= '0' &&
          receivedChars[1] <= '9' &&
          receivedChars[2] >= '0' &&
          receivedChars[2] <= '9' &&
          receivedChars[3] >= '0' &&
          receivedChars[3] <= '9') {
        command_number += 100 * (int(receivedChars[1]) - 48);
        command_number += 10 * (int(receivedChars[2]) - 48);
        command_number += int(receivedChars[3]) - 48;
        if (command_number < 900) {
          silentCommand(command_number, num_chars);
        } else {
          // interactive commands do not have attached data
          if (num_chars != 4) {
            invalidCommandLength(command_number, num_chars);
          } else {
            interactiveCommand(command_number);
          }
        }
      } else {
        commandNotNumeric(); // the next three characters must be numerals
      }
    }
  }
}

// silent commands do not send data to the serial output
// follow the silent command with an interactive command
// if verification of a change is required
void silentCommand(int num, int num_chars) {
  int input_parameter = 0;
  switch (num) {
    case 1:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        startLocalMode();
      }
      break;
    case 2:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        startRemoteMode();
      }
      break;
    case 3:
      if (num_chars != 7) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        if (input_parameter >= 2) {
          if (input_parameter <= 13) {
            digital_pin = input_parameter - 2;
          }
        }
      }
      break;
    case 4:
      if (num_chars != 7) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        if (input_parameter >= 0) {
          if (input_parameter <= 7) {
            analog_pin = input_parameter;
          }
        }
      }
      break;
    case 5:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        pin_locked[digital_pin] = 1;
      }
      break;
    case 6:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        pin_locked[digital_pin] = 0;
      }
      break;
    case 7:
      if (num_chars != 7) {
        invalidCommandLength(num, num_chars);
      } else {
        if (pin_locked[digital_pin] > 0) {
          pinIsLocked();
        } else {
          input_parameter = inputParameterValue(num_chars);
          if (input_parameter >= 0) {
            if (input_parameter <= 5) {
              pin_function[digital_pin] = input_parameter;
              setPinIO(digital_pin, pin_function[digital_pin]);
            }
          }
        }
      }
      break;
    case 8:
      if (num_chars != 7) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        if (input_parameter >= 0) {
          if (input_parameter <= 5) {
            pin_control_src[digital_pin] = input_parameter;
            set_parm_src(digital_pin, pin_parm_src[digital_pin]);
          }
        }
      }
      break;
    case 9:
      if (num_chars != 7) {
        invalidCommandLength(num, num_chars);
      } else {
        if (pin_control_src[digital_pin] > 2) {
          set_parm_src(digital_pin, inputParameterValue(num_chars));
        } else {
          set_parm_src(digital_pin, inputParameterValue(num_chars) - 2);
        }
      }
      break;
    case 10:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_led_times(digital_pin, digital_pin, input_parameter);
      }
      break;
    case 11:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        // interpret remote command and set servo
        input_parameter = inputParameterValue(num_chars);
        set_target(digital_pin, input_parameter);
      }
      break;
    case 12:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_servo_min(digital_pin, input_parameter);
        set_servo_max(digital_pin, pin_servo_max[digital_pin]);
        set_parm_1(digital_pin, pin_parm_1[digital_pin]);
        set_parm_2(digital_pin, pin_parm_2[digital_pin]);
        set_led_times(digital_pin, digital_pin, pin_led_on_time[digital_pin]);
        set_target(digital_pin, pin_servo_target[digital_pin]);
      }
      break;
    case 13:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_servo_max(digital_pin, input_parameter);
        set_servo_min(digital_pin, pin_servo_min[digital_pin]);
        set_parm_1(digital_pin, pin_parm_1[digital_pin]);
        set_parm_2(digital_pin, pin_parm_2[digital_pin]);
        set_led_times(digital_pin, digital_pin, pin_led_on_time[digital_pin]);
        set_target(digital_pin, pin_servo_target[digital_pin]);
      }
      break;
    case 14:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        setSlewRate(input_parameter);
      }
      break;
    case 15:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        setNormalReverse(1);
      }
      break;
    case 16:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        setNormalReverse(0);
      }
      break;
    case 17:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        toggle_d_toggle_state();
      }
      break;
    case 18:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_parm_1(digital_pin, input_parameter);
      }
      break;
    case 19:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_parm_2(digital_pin, input_parameter);
      }
      break;
    case 20:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        analog_in_min[analog_pin] = input_parameter;
        // it should probably never be this close to max and surely less
        if (analog_in_min[analog_pin] > pin_servo_max[analog_pin] - 100) {
          analog_in_min[analog_pin] = pin_servo_max[analog_pin] - 100;
        }
        // but never less than zero
        if (analog_in_min[analog_pin] < 0) {
          analog_in_min[analog_pin] = 0;
        }
        set_analog_range(ain_idx);
      }
      break;
    case 21:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        analog_in_max[analog_pin] = input_parameter;
        // it should probably never be this close to min and surely larger
        if (analog_in_max[analog_pin] < pin_servo_min[analog_pin] + 100) {
          analog_in_max[analog_pin] = pin_servo_min[analog_pin] + 100;
        }
        // this is the most the Nano can do
        if (analog_in_max[analog_pin] > 1023) {
          analog_in_max[analog_pin] = 1023;
        }
        set_analog_range(ain_idx);
      }
      break;
    case 22:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_a_parm_1(ain_idx, input_parameter);
      }
      break;
    case 23:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        set_a_parm_2(ain_idx, input_parameter);
      }
      break;
    case 24:
      if (num_chars != 10) {
        invalidCommandLength(num, num_chars);
      } else {
        input_parameter = inputParameterValue(num_chars);
        if (input_parameter >= 0) {
          if (input_parameter <= 30000) {
            pulse_loop_size = input_parameter;
          }
        }
      }
      break;
    case 25:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        initialize_input_only_mode();
      }
      break;
    case 26:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        initialize_demo_mode();
      }
      break;
    case 27:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        tracker_on = 1;
      }
      break;
    case 28:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        tracker_on = 0;
      }
      break;
    case 29:
      if (num_chars != 4) {
        invalidCommandLength(num, num_chars);
      } else {
        toggle_a_toggle_state(analog_pin);
      }
      break;
    default:
      unimplementedCommand();
      break;
  }
}

//   input only mode
//     the input only mode is intended to assist the developer to identify the
//     associations of pins to hardware in the early stages of development
//     input only mode can be set using a command on the serial interface or
//     a simple change to the code can make module start up in input only mode
//     input only mode does not block commands that can change input pins into
//     output pins or drive an output digital pin high or low but
//     this produces a safe configuration in which many features can still be viewed
//     input pins are relatively safe to implement as long as they are not
//     connected to a power source with a voltage greater or less than the
//     hardware limitations, usually VCC and GROUND.
//     it seems a good idea to verify the association between the pin numbers and
//     the hardware arrangement before you allow any pin to be used as an output
//     use this feature with care!
void initialize_input_only_mode() {
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    pin_function[pin_idx] = 1;
    pin_control_src[pin_idx] = 0;
    pin_parm_src[pin_idx] = 0;
    setPinIO(pin_idx, 1);
  }
}

void toggle_d_toggle_state() {
  if (pin_d_toggle_state[pin_idx] > 0) {
    pin_d_toggle_state[pin_idx] = 0;
  } else {
    pin_d_toggle_state[pin_idx] = 1;
  }
}

void toggle_a_toggle_state(int a_pin) {
  if (pin_a_toggle_state[a_pin] > 0) {
    pin_a_toggle_state[a_pin] = 0;
  } else {
    pin_a_toggle_state[a_pin] = 1;
  }
}

void initialize_demo_mode() {
  // D2/C8 (PWM) Pot by S3 drives servo on C8
  pin_function[0] = 3;  // PWM
  pin_control_src[0] = 3; // logic=POT
  pin_parm_src[0] = 0; // Pot S3 is source
  set_servo_max(0, 1130);
  set_servo_min(0, 1010);

  // D3/C5 (PWM) Pot by S4 drives servo on C5
  pin_function[1] = 3;  // PWM
  pin_control_src[1] = 3; // logic=POT
  pin_parm_src[1] = 2; // Pot S4 is source

  // D4/C4 (PWM) Pot by S1 drives servo on C4
  pin_function[2] = 3;  // PWM
  pin_control_src[2] = 3; // logic=POT
  pin_parm_src[2] = 3; // Pot S1 is source
  set_servo_max(2, 1900);
  set_servo_min(2, 1100);

  // D5/C1 (PWM) Pot by S2 drives servo on C1
  pin_function[3] = 3;  // PWM
  pin_control_src[3] = 3; // logic=POT
  pin_parm_src[3] = 5; // Pot S2 is source
  set_servo_min(3, 1200);

  // The demo has the LED board is connected to Socket D
  for (pin_idx = 4; pin_idx < 12; pin_idx++) {
    pin_function[pin_idx] = 2; // LED
    set_servo_max(pin_idx, 1088);
    set_led_times(pin_idx, pin_idx, 0);
  }

  // D6/LED? (LED) Remote commands drive LED?  (Socket D, pin 10)
  // is this LED 2, 4, 6 or 8?
  pin_control_src[4] = 0; // logic=REM
  pin_parm_src[4] = 0; // not used

  // D7/LED7 (LED) Pot by S4 drives LED7 (Socket D, pin 9)
  pin_control_src[5] = 3; // logic=POT
  pin_parm_src[5] = 2; // Pot S4 is source

  // D8/LED5 (LED) Pot by S3 drives LED5 (Socket D, pin 7)
  pin_control_src[6] = 3; // logic=POT
  pin_parm_src[6] = 0; // Pot S3 is source

  // D9/LED1 (LED) Pot by S1 drives LED1 (Socket D, pin 1)
  pin_control_src[7] = 3; // logic=POT
  pin_parm_src[7] = 3; // Pot S1 is source

  // D10/LED? (LED) Remote commands drive LED?  (Socket D, pin 5)
  // is this LED 2, 4, 6 or 8?
  pin_control_src[8] = 0; // logic=REM
  pin_parm_src[8] = 0; // not used

  // D11/LED? (LED) Remote commands drive LED?  (Socket D, pin 2)
  // is this LED 2, 4, 6 or 8?
  pin_control_src[9] = 0; // logic=REM
  pin_parm_src[9] = 0; // not used

  // D12/LED? (LED) Remote commands drive LED?  (Socket D, pin 8)
  // is this LED 2, 4, 6 or 8?
  pin_control_src[10] = 0; // logic=REM
  pin_parm_src[10] = 0; // not used

  // D13/LED3 (LED) Pot by S2 drives LED3 (Socket D, pin 4)
  pin_control_src[11] = 3; // logic=POT
  pin_parm_src[11] = 5; // Pot S2 is source

  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    setPinIO(pin_idx, pin_function[pin_idx]);
  }
}

void set_analog_range(int a_idx) {
  int range = analog_in_max[a_idx] - analog_in_min[a_idx];
  if (range < 1) {
    range = 1;
  }
  analog_in_range[a_idx] = range;
}

// force pin_parm_src to be within bounds
// pin_parm_src is a digital pin (0-11) if pin_control_src is 1 or 2
// pin_parm_src is an analog pin (0-7) if pin_control_src is 3, 4 or 5
void set_parm_src(int pin_id, int value) {
  pin_parm_src[pin_id] = value;
  switch (pin_control_src[pin_id]) {
    case 1: // TWO-STATE
    case 2: // TRIGGER
      if (value < 0) {
        pin_parm_src[pin_id] = 0;
      }
      if (value > 11) {
        pin_parm_src[pin_id] = 11;
      }
      break;
    case 3: // POT
    case 4: // POT
    case 5: // POT
      if (value < 0) {
        pin_parm_src[pin_id] = 0;
      }
      if (value > 7) {
        pin_parm_src[pin_id] = 7;
      }
      break;
    default:
      pin_parm_src[pin_id] = 0;
      break;
  }
}

void startRemoteMode() { // demo application specific code
  remote_mode = 1;
  if (2 == pin_function[4]) {
    set_led_times(4, 4, 0);
  }
  if (2 == pin_function[3]) {
    set_led_times(3, 3, 0);
  }
}

void startLocalMode() { // demo application specific code
  remote_mode = 0;
  if (2 == pin_function[4]) {
    set_led_times(4, 4, pin_servo_range[4]);
  }
  if (2 == pin_function[3]) {
    set_led_times(3, 3, pin_servo_range[3]);
  }
}

int norm_or_reverse(int pin_id, int value) {
  int range = pin_servo_range[pin_id];
  if (value < 0) {
    value = 0;
  }
  if (value > range) {
    value = range;
  }
  if (pin_reverse_out[pin_id] > 0) {
    return range - value;
  } else {
    return value;
  }
}

// pin_led_on_time and pin_led_off_time are only effective when pin_function = 2 (i.e. LED)
// pin_id is the digital output being controlled
// src_pin is the pin controlling the output if pin_control_src is 1,2,4 or 5
// src_pin = pin_id if pin_control_src is 0 or 3
// value is brightness
void set_led_times(int pin_id, int src_pin, int value) {
  int range = pin_servo_range[src_pin];
  if (value < 0) {
    value = 0;
  }
  if (value > range) {
    value = range;
  }
  if (pin_reverse_out[pin_id] < 1) {
    pin_led_off_time[pin_id] = range - value;
    pin_led_on_time[pin_id] = value;
  } else {
    pin_led_off_time[pin_id] = value;
    pin_led_on_time[pin_id] = range - value;
  }
}

void setNormalReverse(int nr_setting) {
  pin_reverse_out[digital_pin] = nr_setting;
}

void setSlewRate(int slew_rate) {
  if (slew_rate > 0) {
    pin_slew_rate[digital_pin] = slew_rate;
  }
}

// interactive commands respond via serial output
void interactiveCommand(int num) {
  blinkLED6();
  delay(50);
  switch (num) {
    case 999:
      status_report();
      break;
    case 998:
      report_1();
      report_2();
      report_3();
      break;
    case 997:
      report_4();
      break;
    case 996:
      report_1();
      break;
    case 995:
      report_2();
      break;
    case 994:
      report_3();
      break;
    case 993:
      pin_idx = digital_pin;
      report_1_heading();
      report_1_detail();
      Serial.println();
      report_2_heading();
      report_2_detail();
      Serial.println();
      report_3_heading();
      report_3_detail();
      Serial.println();
      break;
    case 992:
      ain_idx = analog_pin;
      report_4_heading();
      report_4_detail();
      Serial.println();
      break;
    case 991:
      print_local();
      break;
    case 990:
      print_user_pin();
      Serial.println();
      break;
    case 989:
      Serial.println(analog_pin);
      break;
    case 988:
      print_locked(digital_pin);
      Serial.println();
      break;
    case 987:
      print_pin_type(digital_pin);
      Serial.println();
      break;
    case 986:
      print_io(pin_io[digital_pin]);
      Serial.println();
      break;
    case 985:
      print_control_src(digital_pin);
      Serial.println();
      break;
    case 984:
      print_parm_src(digital_pin);
      break;
    case 983:
      Serial.println(pin_now[digital_pin]);
      break;
    case 982:
      Serial.println(pin_led_on_time[digital_pin]);
      break;
    case 981:
      Serial.println(pin_led_off_time[digital_pin]);
      break;
    case 980:
      Serial.println(pin_servo_pos[digital_pin]);
      break;
    case 979:
      Serial.println(pin_servo_target[digital_pin]);
      break;
    case 978:
      Serial.println(pin_servo_min[digital_pin]);
      break;
    case 977:
      Serial.println(pin_servo_max[digital_pin]);
      break;
    case 976:
      Serial.println(pin_servo_range[digital_pin]);
      break;
    case 975:
      Serial.print(F("Slew "));
      Serial.println(pin_slew_rate[digital_pin]);
      break;
    case 974:
      printNormalReverse();
      Serial.println();
      break;
    case 973:
      Serial.println(pin_d_toggle_state[digital_pin]);
      break;
    case 972:
      Serial.println(pin_parm_1[digital_pin]);
      break;
    case 971:
      Serial.println(pin_parm_2[digital_pin]);
      break;
    case 970:
      Serial.println(analog_in_value[analog_pin]);
      break;
    case 969:
      Serial.println(analog_in_min[analog_pin]);
      break;
    case 968:
      Serial.println(analog_in_max[analog_pin]);
      break;
    case 967:
      Serial.println(analog_in_range[analog_pin]);
      break;
    case 966:
      Serial.println(analog_in_scaled[analog_pin]);
      break;
    case 965:
      print_pulse_loop_size();
      Serial.println();
      break;
    case 964:
      print_tracker();
      break;
    case 963:
      Serial.println(pin_a_parm_1[digital_pin]);
      break;
    case 962:
      Serial.println(pin_a_parm_2[digital_pin]);
      break;
    default:
      unimplementedCommand();
      break;
  }
}

void blinkLED6() {
  gtSerial.println("wwwwwwwwwwServo ModuleDDDDDDDDD");
}

void printNormalReverse() {
  print_reverse(pin_reverse_out[digital_pin]);
}

int inputParameterValue(int command_len) {
  int parm_len = command_len - 5;
  int val = 0;
  int first_digit = int(receivedChars[5]) - 48;
  int second_digit = int(receivedChars[6]) - 48;
  int third_digit = 0;
  int fourth_digit = 0;
  int fifth_digit = 0;
  if (receivedChars[4] == ',') {
    if (first_digit >= 0) {
      if (first_digit <= 9) {
        if (second_digit >= 0) {
          if (second_digit <= 9) {
            switch (parm_len) {
              case 2:
                val = (10 * first_digit) + second_digit;
                break;
              case 3:
                third_digit = int(receivedChars[7]) - 48;
                if (third_digit >= 0) {
                  if (third_digit <= 9) {
                    val = (100 * first_digit) + (10 * second_digit) + third_digit;
                  }
                }
                break;
              case 5:
                val = (10 * first_digit) + second_digit;
                third_digit = int(receivedChars[7]) - 48;
                fourth_digit = int(receivedChars[8]) - 48;
                fifth_digit = int(receivedChars[9]) - 48;
                if (val < 32) {
                  val = (val * 1000) + (third_digit * 100) + (fourth_digit * 10) + fifth_digit;
                } else {
                  if (32 == val) {
                    val = (third_digit * 100) + (fourth_digit * 10) + fifth_digit;
                    if (val < 768) {
                      val += 32000;
                    } else {
                      val = 0;
                    }
                  } else {
                    val = 0;
                  }
                }
                break;
            }
          }
        }
      }
    }
  }
  return val;
}

void handle_led_on_off() {
  // to prevent visible flicker leds need to be handled both
  //   1. while processing PWM pulses and
  //   2. while doing other processing such as serial input
  // that is why this code is called in two places
  if (2 == pin_function[led_idx]) {
    switch_led_on_off(led_idx);
  }
  led_idx++;
  if (led_idx > pwm_idx_cv) {
    led_idx = 0;
  }
}

void switch_led_on_off(int led_pin) {
  // only affects pins with pin function of 2
  // the user only needs to be concerned with on time and off time
  pin_counter[led_pin]++;
  if (pin_now[led_pin] > 0) {
    if (pin_counter[led_pin] > pin_led_off_time[led_pin]) {
      pin_counter[led_pin] = 0;
      if (pin_led_on_time[led_pin] > 0) {
        pin_now[led_pin] = 0;
        digitalWrite(led_pin + 2, HIGH);
      }
    }
  } else {
    if (pin_counter[led_pin] > pin_led_on_time[led_pin]) {
      pin_counter[led_pin] = 0;
      if (pin_led_off_time[led_pin] > 0) {
        pin_now[led_pin] = 1;
        digitalWrite(led_pin + 2, LOW);
      }
    }
  }
}

// this function affects the duration of the PWM pulses
// pin function of 3 means the pin is being used to control a servo using software PWM
// pin slew rate controls how fast the servo will pan from its current position to its
// target position where it will hold until target has changed
// pin slew rate is set on power up but may be changed under remote control
// the pin slew rate controls the interval between changes in the set point for the indicated servo
// a value of 400 means servo will be nudged once per 4 milliseconds, about 250 times per second, period = 4,000 microseconds
// once per 4 milliseconds = 1000 times in 4 seconds
// each call to servo new pos function changes the servo position (pulse width) by one microsecond
// a change in the servo pulse of 1000 microseconds corresponds with the servo range of 180 degrees (approx.)
// this is 180 degrees in 4 seconds of 90 degrees in 2 seconds (approx.)
// Value 180 degrees 90 degrees 45 degrees
//   400   4.0 secs.  2.0 secs.  1.0 secs.
//   200   2.0 secs.  1.0 secs.  0.5 secs.
//   100   1.0 secs.  0.5 secs.  .25 secs.
//    50   0.5 secs.  .25 secs. .125 secs.
//    10   0.1 secs.  .05 secs. .025 secs.
int adjust_servo_position(int pos, int target, int rate) {
  if (target > pos) {
    if (target < pos + rate) {
      return target;
    } else {
      return pos + rate;
    }
  } else if (target < pos) {
    if (target > pos - rate) {
      return target;
    } else {
      return pos - rate;
    }
  } else {
    return pos;
  }
}

// confine servo min
void set_servo_min(int pin_id, int value) {
  pin_servo_min[pin_id] = value;
  // a pulse width of 100 is way too small but at least it is something
  if (value > pin_servo_max[pin_id] - 100) {
    pin_servo_min[pin_id] = pin_servo_max[pin_id] - 100;
  }
  // the value of 400 is based on the servos we tested with
  // the servos we tested with started to misbehave at about 450
  // if you are not able to get the full range of motion that you expect from
  // your servos changing this value may be one place to start
  if (value < 400) {
    pin_servo_min[pin_id] = 400;
  }
  pin_servo_range[pin_id] = get_range(pin_servo_max[pin_id], pin_servo_min[pin_id]);
}

// confine servo max
void set_servo_max(int pin_id, int value) {
  pin_servo_max[pin_id] = value;
  // a pulse width of 100 is way too small but at least it is something
  if (value < pin_servo_min[pin_id] + 100) {
    pin_servo_max[pin_id] = pin_servo_min[pin_id] + 100;
  }
  if (2 == pin_function[pin_id]) {
    // loosen restrictions only if pin in an LED
    if (value > 32760) {
      pin_servo_max[pin_id] = 32760;
    }
  } else {
    // the value of 3000 is based on the servos we tested with
    // the servos we tested with started to misbehave at about 2500
    // if you are not able to get the full range of motion that you expect from
    // your servos changing this value may be one place to start
    if (value > 3000) {
      pin_servo_max[pin_id] = 3000;
    }
  }
  pin_servo_range[pin_id] = get_range(pin_servo_max[pin_id], pin_servo_min[pin_id]);
}

int get_range(int max_val, int min_val) {
  int range = max_val - min_val;
  if (range < 1) {
    range = 1;
  }
  return range;
}

// confine parm_1
void set_parm_1(int pin_id, int value) {
  pin_parm_1[pin_id] = value;
  if (value < 0) {
    pin_parm_1[pin_id] = 0;
  }
  if (value > pin_servo_range[pin_id]) {
    pin_parm_1[pin_id] = pin_servo_range[pin_id];
  }
}

// confine parm_2
void set_parm_2(int pin_id, int value) {
  pin_parm_2[pin_id] = value;
  if (value < 0) {
    pin_parm_2[pin_id] = 0;
  }
  if (value > pin_servo_range[pin_id]) {
    pin_parm_2[pin_id] = pin_servo_range[pin_id];
  }
}

// confine a_parm_1
void set_a_parm_1(int pin_id, int value) {
  pin_a_parm_1[pin_id] = value;
  if (value < 0) {
    pin_a_parm_1[pin_id] = 0;
  }
  if (value > pin_servo_range[pin_id]) {
    pin_a_parm_1[pin_id] = pin_servo_range[pin_id];
  }
}

// confine a_parm_2
void set_a_parm_2(int pin_id, int value) {
  pin_a_parm_2[pin_id] = value;
  if (value < 0) {
    pin_a_parm_2[pin_id] = 0;
  }
  if (value > pin_servo_range[pin_id]) {
    pin_a_parm_2[pin_id] = pin_servo_range[pin_id];
  }
}

// confine servo target to its range
void set_target(int pin_id, int value) {
  pin_servo_target[pin_id] = value;
  if (value < pin_servo_min[pin_id]) {
    pin_servo_target[pin_id] = pin_servo_min[pin_id];
  }
  if (value > pin_servo_max[pin_id]) {
    pin_servo_target[pin_id] = pin_servo_max[pin_id];
  }
}

void setPinIO(int idx, int pin_fnc) {
  // idx of 4 is pin D6, etc.
  bool ok2set = true;
  if (idx == 4) ok2set = false; // don't mess with extra serial RX
  if (idx == 10) ok2set = false; // don't mess with extra serial TX
  //if (idx == 1) ok2set = false; // don't mess with Arduino PWM on D3
  if (ok2set) {
    pin_io[idx] = 0;
    pinMode(idx + 2, INPUT);
    if (pin_fnc > 1) {
      pin_io[idx] = 1;
      pinMode(idx + 2, OUTPUT);
      digitalWrite(idx + 2, LOW);
    }
  }
}

void invalidModuleCode() {
  Serial.println(F("S1 is the module code"));
}

void unimplementedCommand() {
  Serial.println(F("Command is not implemented"));
}

void commandNotNumeric() {
  Serial.println(F("Command is a 3-digit number"));
}

void invalidCommandLength(int num, int len) {
  Serial.print(F("For command number "));
  Serial.print(num);
  Serial.print(F(", a length of "));
  Serial.print(len);
  Serial.println(F(" characters (including data) is invalid"));
}

void pinIsLocked() {
  Serial.print(F("Pin "));
  print_user_pin();
  Serial.println(F(" is locked"));
}

void status_report() {
  Serial.println(F("Servo Module State"));
  Serial.println(F("------------------"));
  delay(100);
  Serial.println(F("(RX and TX are not shown)"));
  Serial.print(F("remote/local: "));
  print_local();
  delay(100);
  Serial.print(F("PWM: "));
  print_tracker();
  delay(100);
  Serial.print(F("current digital pin: "));
  print_user_pin();
  Serial.println();
  Serial.print(F("current analog pin: A"));
  Serial.println(analog_pin);
  delay(100);
  print_pulse_loop_size();
  Serial.println();
  Serial.println();
  delay(100);

  report_1();
  report_2();
  report_3();
  report_4();

  Serial.print(F("Last command received: S"));
  Serial.println(receivedChars);
  num_chars = 0;
  for (i = 0; i < 32; i++) {
    if (receivedChars[i] == '\0') {
      num_chars = i;
      i = 32;
    }
  }
  for (i = 0; i < num_chars; i++) {
    Serial.print(F("receivedChars[ "));
    Serial.print(i);
    Serial.print(F(" ] = "));
    Serial.print(receivedChars[i]);
    Serial.print(F(", int val = "));
    Serial.println(int(receivedChars[i]));
    delay(100);
  }
}

void print_user_pin() {
  Serial.print(F("D"));
  Serial.print(digital_pin + 2);
}

void print_pulse_loop_size() {
  Serial.print(F("PLS "));
  Serial.print(pulse_loop_size);
}

void report_1() {
  report_1_heading();
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    report_1_detail();
  }
  Serial.println();
}

void report_1_heading() {
  Serial.println(F("Pin  Read  Function  SWI/LOC  I/O  Ctrl Src  Direction"));
  delay(100);
  Serial.println(F("---  ----  --------  -------  ---  --------  ---------"));
  delay(100);
}

void report_1_detail() {
  if (pin_idx < 8) {
    Serial.print(F(" "));
  }
  Serial.print(F("D"));
  Serial.print(pin_idx + 2);
  Serial.print(F("     "));
  Serial.print(pin_state[pin_idx]);
  Serial.print(F("     "));
  print_pin_type(pin_idx);
  Serial.print(F("      "));
  print_locked(pin_idx);
  Serial.print(F("    "));
  print_io(pin_io[pin_idx]);
  Serial.print(F("  "));
  print_control_src(pin_idx);
  Serial.print(F("   "));
  print_reverse(pin_reverse_out[pin_idx]);
  Serial.println();
  delay(100);
}

void report_2() {
  report_2_heading();
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    report_2_detail();
  }
  Serial.println();
}

void report_2_heading() {
  Serial.println(F("Pin   State  On-time  Off-time  D-Var-1  D-Var-2"));
  delay(100);
  Serial.println(F("---   -----  -------  --------  -------  -------"));
  delay(100);
}

void report_2_detail() {
  if (pin_idx < 8) {
    Serial.print(F(" "));
  }
  Serial.print(F("D"));
  Serial.print(pin_idx + 2);
  Serial.print(F("   "));
  if (pin_function[pin_idx] < 2) {
    Serial.print(F("IN  "));
    Serial.print(pin_now[pin_idx]);
  } else {
    Serial.print(F("OUT  "));
  }
  Serial.print(F("   "));
  print_num(pin_led_on_time[pin_idx]);
  Serial.print(F("     "));
  print_num(pin_led_off_time[pin_idx]);
  Serial.print(F("    "));
  print_num(pin_parm_1[pin_idx]);
  Serial.print(F("    "));
  print_num(pin_parm_2[pin_idx]);
  Serial.println();
  delay(100);
}

void report_3() {
  report_3_heading();
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    report_3_detail();
  }
  Serial.println();
}

void report_3_heading() {
  Serial.println(F("Pin    Pos   Target    Min    Max    Range  Slew Rate"));
  delay(100);
  Serial.println(F("---    ---   ------    ---    ---    -----  ---------"));
  delay(100);
}

void report_3_detail() {
  if (pin_idx < 8) {
    Serial.print(F(" "));
  }
  Serial.print(F("D"));
  Serial.print(pin_idx + 2);
  Serial.print(F("  "));
  print_num(pin_servo_pos[pin_idx]);
  Serial.print(F("   "));
  print_num(pin_servo_target[pin_idx]);
  Serial.print(F("   "));
  print_num(pin_servo_min[pin_idx]);
  Serial.print(F("  "));
  print_num(pin_servo_max[pin_idx]);
  Serial.print(F("    "));
  print_num(pin_servo_range[pin_idx]);
  Serial.print(F("   "));
  print_num(pin_slew_rate[pin_idx]);
  Serial.println();
  delay(100);
}

void report_4() {
  report_4_heading();
  for (ain_idx = 0; ain_idx < 8; ain_idx++) {
    report_4_detail();
  }
  Serial.println();
}

void report_4_heading() {
  Serial.println(F("Ain   Pin  Dv   Value    Min    Max   Range    Scaled A-Var-1 A-Var-2"));
  delay(100);
  Serial.println(F("---   ---  --   -----    ---    ---   -----    ------ ------- -------"));
  delay(100);
}

void report_4_detail() {
  Serial.print(F(" A"));
  Serial.print(ain_idx);
  Serial.print(F(" "));
  print_num(analog_in_pin[ain_idx]);
  Serial.print(F("   "));
  Serial.print(pin_a_now[ain_idx]);
  Serial.print(F("   "));
  print_num(analog_in_value[ain_idx]);
  Serial.print(F("  "));
  print_num(analog_in_min[ain_idx]);
  Serial.print(F("  "));
  print_num(analog_in_max[ain_idx]);
  Serial.print(F("   "));
  print_num(analog_in_range[ain_idx]);
  Serial.print(F("     "));
  print_num(analog_in_scaled[ain_idx]);
  Serial.print(F("   "));
  print_num(pin_a_parm_1[pin_idx]);
  Serial.print(F("   "));
  print_num(pin_a_parm_2[pin_idx]);
  Serial.println();
  delay(100);
}

void print_num(int value) {
  if (value < 10000) {
    Serial.print(F(" "));
  }
  if (value < 1000) {
    Serial.print(F(" "));
  }
  if (value < 100) {
    Serial.print(F(" "));
  }
  if (value < 10) {
    Serial.print(F(" "));
  }
  Serial.print(value);
}

void print_local() {
  if (remote_mode > 0) {
    Serial.println(F("REMOTE"));
  } else {
    Serial.println(F("LOCAL"));
  }
}

void print_tracker() {
  Serial.print(F("TRACKER "));
  if (tracker_on > 0) {
    Serial.println(F("ON"));
  } else {
    Serial.println(F("OFF"));
  }
}

void print_reverse(int state) {
  if (state > 0) {
    Serial.print(F("REVERSE"));
  } else {
    Serial.print(F("NORMAL"));
  }
}

void print_pin_type(int index) {
  switch (pin_function[index]) {
    case 0:
      Serial.print(F("RAW"));
      break;
    case 1:
      Serial.print(F("DEB"));
      break;
    case 2:
      Serial.print(F("LED"));
      break;
    case 3:
      Serial.print(F("PWM"));
      break;
    case 4:
      Serial.print(F("ARP"));
      break;
    default:
      Serial.print(F("UND"));
      break;
  }
}

void print_locked(int index) {
  if (pin_locked[index] > 0) {
    Serial.print(F("LOC"));
  } else {
    Serial.print(F("UNL"));
  }
}

void print_io(int state) {
  if (state > 0) {
    Serial.print(F("OUT"));
  } else {
    Serial.print(F("IN "));
  }
}

void print_control_src(int index) {
  if (pin_function[pin_idx] < 2) {
    Serial.print(F("INPUT   "));
  } else {
    switch (pin_control_src[index]) {
      case 0:
        Serial.print(F("REM    "));
        break;
      case 1:
        Serial.print(F("BTN   D"));
        Serial.print(pin_parm_src[index] + 2);
        break;
      case 2:
        Serial.print(F("TGR   D"));
        Serial.print(pin_parm_src[index] + 2);
        break;
      case 3:
        Serial.print(F("POT   A"));
        Serial.print(pin_parm_src[index]);
        break;
      case 4:
        Serial.print(F("POTB  A"));
        Serial.print(pin_parm_src[index]);
        break;
      case 5:
        Serial.print(F("POTT  A"));
        Serial.print(pin_parm_src[index]);
        break;
    }
  }
}

void print_parm_src(int index) {
  if (pin_control_src[index] > 2) {
    Serial.println(pin_parm_src[index]);
  } else {
    Serial.println(pin_parm_src[index] + 2);
  }
}
